/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <string>
#include <vector>
#include <complex>
#include <math.h>
#include <algorithm>

#include <flowdesigner/BufferedNode.h>
#include <flowdesigner/Buffer.h>
#include <flowdesigner/Vector.h>
#include <flowdesigner/FDComplex.h>
#include <flowdesigner/Matrix.h>

#include <hark/Map.h>
#include <hark/Source.h>

#include <boost/python.hpp>
#include <boost/python/numeric.hpp>
#include <boost/python/suite/indexing/vector_indexing_suite.hpp>
#include <boost/python/suite/indexing/map_indexing_suite.hpp>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/uuid/uuid_generators.hpp>

namespace bp = boost::python;
namespace bu = boost::uuids;
using namespace FD;
using namespace std;

class PyCodeExecutor;


DECLARE_NODE(PyCodeExecutor);
/*Node
 *
 * @name PyCodeExecutor
 * @category HARK:Python
 * @description This node executs an external python code. 
 *
 * @parameter_name DIRECTORY_NAME
 * @parameter_type string
 * @parameter_description The directory name of your python code. It will inserted to sys.path.
 *
 * @parameter_name MODULENAME
 * @parameter_type string
 * @parameter_value samplecode
 * @parameter_description Your python module name to import, i.e., your python code file name WITHOUT extension.
 * 
 * @parameter_name CLASSNAME
 * @parameter_type string
 * @parameter_value HarkNode
 * @parameter_description Your class name to call in this node.
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 * 
 END*/

class PyCodeExecutor : public BufferedNode {
  vector<int> inputsID;
  vector<string> inputsName; 

  vector<int> outputsID;
  vector<string> outputsName;
  vector<string> outputsType;

  bp::object main_namespace;
  string myNodeName;
  string classname;
  string filename;

  bool isDebug;
  bool isCalculated;
  string uuid;

  void uuidreplace(string& str, const string& from, const string& to){
    string::size_type pos = 0;
    while(pos = str.find(from, pos), pos != string::npos){
      str.replace(pos, from.length(), to);
      pos += to.length();
    }
  }

public:
  PyCodeExecutor(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    // get parameters
    string dirname = "";
    if(parameters.exist("DIRECTORY_NAME"))
      dirname = object_cast<string>(parameters.get("DIRECTORY_NAME"));
    filename = object_cast<string>(parameters.get("MODULENAME"));
    classname = object_cast<string>(parameters.get("CLASSNAME"));
    isDebug = dereference_cast<bool>(parameters.get("DEBUG"));
    isCalculated = false;

    // generate unique id
    uuid = boost::lexical_cast<string>(bu::random_generator()()); 
    uuidreplace(uuid, "-", "_");
    cout << uuid << endl;

    Py_Initialize();
    try{
      bp::object main_module = bp::import("__main__");
      main_namespace = main_module.attr("__dict__");
      bp::numeric::array::set_module_and_type("numpy", "ndarray");
      
      // get nodeID and name
      myNodeName = nodeName + uuid;
      int nodeID = -1;
      stringstream ss(nodeName.substr(nodeName.rfind("_")+1));
      ss >> nodeID;
      main_namespace["numpy"] = bp::import("numpy");    
      
      // add the path for user's script
      main_namespace["sys"] = bp::import("sys");
      if(dirname.length() > 1){
      	bp::list path = bp::extract<bp::list>(main_namespace["sys"].attr("path"));
      	path.append(dirname.c_str());
      }
      exec("from harkpython import *", main_namespace);
      main_namespace[filename.c_str()] = bp::import(filename.c_str());

      // instantiate user's code.
      main_namespace[myNodeName.c_str()] 
      	= main_namespace[filename.c_str()].attr(classname.c_str())();
      main_namespace[myNodeName.c_str()].attr("nodeID") = nodeID;
      main_namespace[myNodeName.c_str()].attr("nodeName") = nodeName;
      main_namespace[myNodeName.c_str()].attr("makeup")();

      // tell python interpreter how to deal with STL obects.
      if (main_namespace.attr("has_key")("convertersRegistered") == false){
        bp::class_<vector<complex<float> > > ("PyVecComplexFloat")
          .def(bp::vector_indexing_suite<vector<complex<float> > >());
        bp::class_<vector<int> > ("PyVecInt")
          .def(bp::vector_indexing_suite<vector<int> >());
        bp::class_<vector<float> > ("PyVecFloat")
          .def(bp::vector_indexing_suite<vector<float> >());
        bp::class_<map<int, vector<int> > > ("PyMapVecInt")
          .def(bp::map_indexing_suite<map<int, vector<int> > >());
        bp::class_<map<int, vector<float> > > ("PyMapVecFloat")
          .def(bp::map_indexing_suite<map<int, vector<float> > >());
        bp::class_<map<int, vector<complex<float> > > > ("PyMapVecComplex")
          .def(bp::map_indexing_suite<map<int, vector<complex<float> > > >());
        bp::class_<map<int, float > > ("PyMapPrimFloat")
        .def(bp::map_indexing_suite<map<int, float > >());

        bp::to_python_converter<Vector<int>, Vector2py<int> >();
        bp::to_python_converter<Vector<float>, Vector2py<float> >();
        bp::to_python_converter<Vector<complex<float> >, Vector2py<complex< float> > >();
        bp::to_python_converter<Matrix<int>, Matrix2py<int> >();
        bp::to_python_converter<Matrix<float>, Matrix2py<float> >();
        bp::to_python_converter<Matrix<complex<float> >, Matrix2py<complex<float> > >();
        bp::to_python_converter<Int, Prim2py<int> >();
        bp::to_python_converter<Float, Prim2py<float> >();
        bp::to_python_converter<Complex<float>, ComplexPrim2py<float> >();
        bp::to_python_converter<Complex<double>, ComplexPrim2py<double> >();
        bp::to_python_converter<Complex<int>, ComplexPrim2py<int> >();
        bp::to_python_converter<Source, Source2py>();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<int> > > >, Map2py<int> >();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<float> > > >, Map2py<float> >();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<complex<float> > > > >, Map2py<complex<float> > >();

        // for PARAM_MAP
        //bp::to_python_converter<RCPtr< Map<string, RCPtr< Map <string, ObjectRef> > > >, MapMap2py >();

        main_namespace["convertersRegistered"] = true;
      }
    } catch(bp::error_already_set){
      PyErr_Print();
      throw (new NodeException(NULL, "Error in constructor", __FILE__, __LINE__));
    }
    inOrder = true;
  }

  template <typename T>
  struct Map2py{
    static PyObject* convert(RCPtr<Map<int, RCPtr<Vector<T> > > > const& s){
      bp::dict topy = bp::dict();
      typename Map<int, RCPtr<Vector<T> > >::const_iterator it = s->begin();
      typename Map<int, RCPtr<Vector<T> > >::const_iterator ed = s->end();
      for(; it != ed; it++) topy[it->first] = object_cast<Vector<T> >((it->second));
      return bp::incref(bp::object(topy).ptr());
    }
  };
/*
  struct MapMap2py{
    static PyObject* convert(RCPtr< Map<string, RCPtr< Map <string, ObjectRef> > > > const& s){
      bp::dict topy = bp::dict();
      typename Map<string, RCPtr< Map <string, ObjectRef> > >::const_iterator it1 = s->begin();
      typename Map<string, RCPtr< Map <string, ObjectRef> > >::const_iterator ed1 = s->end();
      for(; it1 != ed1; it1++){
        bp::dict map1 = bp::dict();
        typename  Map <string, ObjectRef>::const_iterator it2 = (it1->second)->begin();
        typename  Map <string, ObjectRef>::const_iterator ed2 = (it1->second)->end();
        for(; it2 != ed2; it2++){
            if(typeid(Int) == typeid(*(it2->second))){
              RCPtr<Int> tmp = it2->second;
              map1[it2->first] = *tmp;
            } else if(typeid(Float) == typeid(*(it2->second))){
              RCPtr<Float> tmp = it2->second;
              map1[it2->first] = *tmp;
            } else if(typeid(String) == typeid(*(it2->second))){
              RCPtr<String> tmp = it2->second;
              map1[it2->first] = *tmp;
            }
        }

        topy[it1->first] = map1;
      }
      return bp::incref(bp::object(topy).ptr());
    }
  };
*/
  template <typename T>
  struct Prim2py{
    static PyObject* convert(NetCType<T> const& s) {
      NetCType<T> &topy = const_cast< NetCType < T >  & > ( s );
      return bp::incref(bp::object(topy.val()).ptr());
    }
  };
  template <typename T>
  struct ComplexPrim2py{
    static PyObject* convert(Complex<T> const& s){
      Complex<T> &topy = const_cast<Complex<T> &> (s);
      return bp::incref(bp::object(topy.val()).ptr());
    }
  };
  template <typename T>
  struct Vector2py{
    static PyObject* convert(Vector<T> const& s){
      return bp::incref(bp::object(vector<T>(s)).ptr());
    }
  };

  template <typename T>
  struct Matrix2py{
    static PyObject* convert(Matrix<T> const& input){
      vector<T> data(input.nrows() * input.ncols());
      for(int row=0; row<input.nrows(); row++)
	for(int col=0; col<input.ncols(); col++)
	  data[row * input.ncols() + col] = input(row, col);

      bp::numeric::array topy(data);
      topy.resize(bp::make_tuple(input.nrows(), input.ncols()));

      return bp::incref(bp::object(topy).ptr());
    }
  };


  struct Source2py{
    static PyObject* convert(Source const& s){
      bp::dict topy = bp::dict();
      Source& src = const_cast<Source&> (s);
      topy["id"] = src.id;
      topy["x"] = bp::list();
      topy["x"].attr("append")(src.x[0]);
      topy["x"].attr("append")(src.x[1]);
      topy["x"].attr("append")(src.x[2]);
      topy["power"] = src.power;
      return bp::incref(topy.ptr());	
    }
  };
  ~PyCodeExecutor(){
    try{
      main_namespace[myNodeName.c_str()].attr("__del__")();
    } catch(bp::error_already_set){
      if(isCalculated){
      	PyErr_Print();
      	throw(new NodeException(NULL, "Error in destructor", __FILE__, __LINE__));      
      } else{}
    }
  }
  
  virtual int translateInput(string inputName){
    int id = addInput(inputName);
    inputsName.push_back(inputName);
    inputsID.push_back(id);

    if (isDebug)
      cout << "translate Input (" << myNodeName << ":" << classname << ") "
      	   << inputName << " (ID: " << id << ")" << endl;
    return id;
  }
  
  virtual int translateOutput(string outputName){
    vector<string>::iterator result 
      = find(outputsName.begin(), outputsName.end(), outputName);
    
    if (result != outputsName.end()){
      if(isDebug)
      	cout << outputName << " already exists: " 
      	     << result - outputsName.begin() << endl;
      return outputsID[result - outputsName.begin()];
    }
    int id = addOutput(outputName);
    outputsName.push_back(outputName);
    outputsID.push_back(id);

    if (isDebug)
      cout << "translate Output: (" << myNodeName << "::" << classname << ")" 
    	     << " Name: " << outputName << " ID: " << id << endl;
    bp::tuple nam;
    bp::tuple typ;
    try{
      nam = bp::extract<bp::tuple>(main_namespace[myNodeName.c_str()].attr("outputNames"));
      typ = bp::extract<bp::tuple>(main_namespace[myNodeName.c_str()].attr("outputTypes"));
    } catch(bp::error_already_set){
      PyErr_Print();
      stringstream ss;
      ss << "Error during translate Output (" 
      	 << filename << "::" << myNodeName << "::" << classname << ")" 
      	 << "outputNames or outputTypes is not defined as tuple" << endl;
      throw (new NodeException(NULL, ss.str(),  __FILE__, __LINE__));
    }

    int inam = 0;
    try{
      inam = bp::extract<int>(nam.attr("index")(outputName));
    } catch(bp::error_already_set){
      PyErr_Print();
      stringstream ss;
      ss << "Error during translate Output (" 
      	 << filename << "::" << myNodeName << "::" << classname << ")" 
      	 << " Output Name: " << outputName << " of ID: " << id
      	 << " is not found in outputNames." << endl
      	 << "candidates are:::" << endl;
      for(int i=0; i<bp::len(nam); i++){
      	string str =  bp::extract<string>(nam[i]);
      	ss << str << endl;
      }
      throw (new NodeException(NULL, ss.str(),  __FILE__, __LINE__));
    }
    outputsType.push_back((string) bp::extract<string>(typ[inam]));

    return id;
  }
  
  virtual void initialize(void)
  {
    vector<int>::iterator id;

    int maxahead = 0, maxback = 0;
    for(id = outputsID.begin(); id != outputsID.end(); id++){
      if(maxahead < outputs[*id].lookAhead) maxahead = outputs[*id].lookAhead;
      if(maxback < outputs[*id].lookBack) maxback = outputs[*id].lookBack;
    }
    for(id = outputsID.begin(); id != outputsID.end(); id++){
      outputs[*id].lookAhead = maxahead + 1;
      outputs[*id].lookBack = maxback + 1;
    }
    this->BufferedNode::initialize();
  }

  template<typename T>
  void givePython(RCPtr<T> input, string inputName){
    main_namespace[myNodeName.c_str()].attr(inputName.c_str()) = *input;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // get input
    vector<string>::iterator inputsNameit = inputsName.begin();
    vector<int>::iterator inputsIDit = inputsID.begin();
    try{
      for(; inputsIDit < inputsID.end(); inputsIDit++, inputsNameit++){
        ObjectRef input = getInput(*inputsIDit, count);
        if(isDebug)
          cout << "Input terminal (" << myNodeName << "::" << classname << ")  "
               << "Name: " << *inputsNameit << ", ID: " << *inputsIDit << endl;

        // give data to python depending on their types 
        if(typeid(Int) == typeid(*input))
          givePython(RCPtr<Int>(input), *inputsNameit);
        else if(typeid(Float) == typeid(*input))
          givePython(RCPtr<Float>(input), *inputsNameit);
        else if(typeid(Complex<float>) == typeid(*input))
          givePython(RCPtr<Complex<float> >(input), *inputsNameit);
        else if(typeid(Source) == typeid(*input))
          givePython(RCPtr<Source> (input), *inputsNameit);
        else if(typeid(Vector<int>) == typeid(*input))
          givePython(RCPtr<Vector<int> >(input), *inputsNameit);	
        else if(typeid(Vector<float>) == typeid(*input))
          givePython(RCPtr<Vector<float> >(input), *inputsNameit);
        else if(typeid(Vector<complex<float> >) == typeid(*input))
          givePython(RCPtr<Vector<complex<float> > >(input), *inputsNameit);	
        else if(typeid(Matrix<float>) == typeid(*input))
          givePython(RCPtr<Matrix<float> >(input), *inputsNameit);
        else if(typeid(Matrix<int>) == typeid(*input))
          givePython(RCPtr<Matrix<int> >(input), *inputsNameit);
        else if(typeid(Matrix<complex<float> >) == typeid(*input))
          givePython(RCPtr<Matrix<complex<float> > >(input), *inputsNameit);
        else if(typeid(Vector<ObjectRef>) == typeid(*input)){
          Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> > (input);
          bp::list topylist = bp::list();
          for(unsigned int i = 0; i < vec.size(); i++){
            if(typeid(Source) == typeid(*(vec[i])))
              topylist.append(object_cast<Source>(vec[i]));
            else{
              stringstream ss;
              ss << "Input of Vector<ObjectRef> [ObjectRef = " << typeid(*(vec[i])).name() << "] is not supported";
              throw new NodeException(NULL, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topylist;
        }
        else if(typeid(*input) == typeid(Map<int, ObjectRef>)){
          Map<int, ObjectRef>  topy = object_cast<Map<int, ObjectRef> > (input);
          bp::dict topydict = bp::dict();

          Map<int, ObjectRef >::iterator it = topy.begin();
          for(; it != topy.end(); it++){
            if(typeid(*(it->second)) == typeid(Vector<int>))
              topydict[it->first] = object_cast<Vector<int> > (it->second);
            else if(typeid(*(it->second)) == typeid(Vector<float>))
              topydict[it->first] = object_cast<Vector<float> > (it->second);
            else if(typeid(*(it->second)) == typeid(Vector<complex<float> >))
              topydict[it->first] = object_cast<Vector<complex<float> > > (it->second);
            else if(typeid(*(it->second)) == typeid(Float))
              topydict[it->first] = object_cast<Float> (it->second);
            else if(typeid(*(it->second)) == typeid(Matrix<float> ))
              topydict[it->first] = object_cast<Matrix<float> > (it->second);
            else if(typeid(*(it->second)) == typeid(Matrix<int> ))
              topydict[it->first] = object_cast<Matrix<int> > (it->second);
            else if(typeid(*(it->second)) == typeid(Matrix<complex<float> > ))
              topydict[it->first] = object_cast<Matrix<complex<float> > > (it->second);
            else{
              stringstream ss;
              ss << "Map value" << typeid(*(it->second)).name() << " is not supported";
              throw new NodeException(NULL, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topydict;
        }
        else if(typeid(*input) == typeid(Map<string, ObjectRef>)){
          Map<string, ObjectRef>  topy = object_cast<Map<string, ObjectRef> > (input);
          bp::dict topydict = bp::dict();

          Map<string, ObjectRef >::iterator it = topy.begin();
          for(; it != topy.end(); it++){
            if(typeid(*(it->second)) == typeid(Map<string, int>))
              topydict[it->first] = object_cast<Int> (it->second);
            else if(typeid(*(it->second)) == typeid(Map<string, float>))
              topydict[it->first] = object_cast<Int> (it->second);
            else if(typeid(*(it->second)) == typeid(Map<string, string>))
              topydict[it->first] = object_cast<String> (it->second);
            else{
              stringstream ss;
              ss << "Map<string, T> value" << typeid(*(it->second)).name() << " is not supported";
              throw new NodeException(NULL, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topydict; 
        }
        else{
          if(isDebug) cout << endl;
          stringstream ss;
          ss << "Type (" << typeid(*input).name() << ") of input terminal "
             << *inputsNameit << " is unsupported." << endl;
          throw new NodeException (NULL, ss.str(), __FILE__, __LINE__); 
        }
      }
    } catch(bp::error_already_set){
      PyErr_Print();
      throw (new NodeException(NULL, "Error during getting inputs", __FILE__, __LINE__));
    }
    
    if(isDebug)
      cerr << "calculating..." << endl;
    //calculate
    try{
      main_namespace[myNodeName.c_str()].attr("count") = count;
      exec((myNodeName + ".calculate()").c_str(), main_namespace);
    }
    catch(bp::error_already_set){
      PyErr_Print();
      throw (new NodeException(NULL, "Error during calculate function execution",
			       __FILE__, __LINE__));      
    }
    
    //set output
    vector<string>::iterator outputsNameit = outputsName.begin();
    vector<int>::iterator outputsIDit = outputsID.begin();
    vector<string>::iterator outputsTypeit = outputsType.begin();
    try{
      main_namespace[myNodeName.c_str()].attr("convertValues")();
    } catch(bp::error_already_set){
      PyErr_Print();
      throw (new NodeException(NULL, "converting values failed",  __FILE__, __LINE__));      
    }
    try{
      bp::dict d = bp::extract<bp::dict>	(main_namespace[myNodeName.c_str()].attr("outputValues"));
      
      for(; outputsIDit < outputsID.end(); outputsTypeit++, outputsNameit++, outputsIDit++){
	
	if(isDebug)
	  cout << "Output terminal (" << myNodeName << "::" << classname << ")  "
	       << "Name: " << *outputsNameit 
	       << ", Type: " << *outputsTypeit
	       << ", ID: " << *outputsIDit << endl;

	if(*outputsTypeit == "prim_float"){
	  (*(outputs[*outputsIDit].buffer))[count] = 
	    (new Float(bp::extract<float>(d.get(*outputsNameit))));;
	}
	else if(*outputsTypeit == "prim_int"){
	  (*(outputs[*outputsIDit].buffer))[count] = 
	    (new Int(bp::extract<int>(d.get(*outputsNameit))));
	}
	else if(*outputsTypeit == "prim_bool"){
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Bool(bp::extract<bool>(d.get(*outputsNameit))));
	}
	else if(*outputsTypeit == "prim_complex"){
	  (*(outputs[*outputsIDit].buffer))[count] = 
	    (new Complex<float >(bp::extract<complex<float> >(d.get(*outputsNameit))));
	}
	else if(*outputsTypeit == "prim_bool"){
	  (*(outputs[*outputsIDit].buffer))[count] = 
	    (new Bool(bp::extract<bool>(d.get(*outputsNameit))));
	}
	else if(*outputsTypeit == "prim_source"){
	  RCPtr<Source> src = RCPtr<Source>(new Source);
	  bp::dict srcd = bp::extract<bp::dict>(d.get(*outputsNameit));
	  src->id = bp::extract<int>(srcd.get("id"));
	  src->x[0] = bp::extract<float>(srcd.get("x")[0]);
	  src->x[1] = bp::extract<float>(srcd.get("x")[1]);
	  src->x[2] = bp::extract<float>(srcd.get("x")[2]);
	  src->power = bp::extract<float>(srcd.get("power"));
	  (*(outputs[*outputsIDit].buffer))[count] = src;
	}
	else if(*outputsTypeit == "vector_int"){
	  bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Vector<int> > frompy(new Vector<int>());
	  for(int i=0; i<bp::len(tmplist); i++)
	    (*frompy).push_back((int)bp::extract<int>(tmplist[i]));
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "vector_float"){
	  bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Vector<float> > frompy(new Vector<float>());
	  for(int i=0; i<len(tmplist); i++)
	    (*frompy).push_back(bp::extract<float>(tmplist[i]));
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "vector_complex"){
	  bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Vector<complex<float> > > frompy(new Vector<complex<float> >());
	  for(int i=0; i<len(tmplist); i++)
	    (*frompy).push_back(bp::extract<complex<float> >(tmplist[i]));
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "matrix_int"){
	  bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Matrix<int > > frompy 
	    (new Matrix<int >(bp::len(tmpmat), bp::len(tmpmat[0])));
	  for(int r=0; r<bp::len(tmpmat); r++)
	    for(int c=0; c<bp::len(tmpmat[0]); c++)
	      (*frompy)(r, c) = bp::extract<int>(tmpmat[r][c]);
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}  
	else if(*outputsTypeit == "matrix_float"){
	  bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Matrix<float> > frompy (new Matrix<float>(len(tmpmat), len(tmpmat[0])));
	  for(int r=0; r<len(tmpmat); r++)
	    for(int c=0; c<len(tmpmat[0]); c++)
	      (*frompy)(r, c) = bp::extract<float>(tmpmat[r][c]);
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}  
	else if(*outputsTypeit == "matrix_complex"){
	  bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Matrix<complex<float> > > frompy (new Matrix<complex<float> >(len(tmpmat), len(tmpmat[0])));
	  for(int r=0; r<len(tmpmat); r++)
	    for(int c=0; c<len(tmpmat[0]); c++)
	      (*frompy)(r, c) = bp::extract<complex<float> >(tmpmat[r][c]);
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "vector_source"){
	  bp::list tmpvec = bp::extract<bp::list>(d.get(*outputsNameit));
	  RCPtr<Vector<ObjectRef> > frompy (new Vector<ObjectRef>());
	  for(int i=0; i<bp::len(tmpvec); i++){
	    RCPtr<Source> src = RCPtr<Source>(new Source);
	    bp::dict srcd = bp::extract<bp::dict>(tmpvec[i]);
	    src->id = bp::extract<int>(srcd.get("id"));
	    src->x[0] = bp::extract<float>(srcd.get("x")[0]);
	    src->x[1] = bp::extract<float>(srcd.get("x")[1]);
	    src->x[2] = bp::extract<float>(srcd.get("x")[2]);
	    src->power = bp::extract<float>(srcd.get("power"));
	    (*frompy).push_back(src);
	  }
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "map_vector_int"){
	  bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
	  bp::list keys = bp::extract<bp::list>(map.attr("keys")());
	  RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);

	  for(int i=0; i<bp::len(keys); i++){
	    RCPtr<Vector<int> > value = RCPtr<Vector<int> >(new Vector<int>);
	    bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
	    for(int j=0; j<bp::len(tmplist); j++)
	      (*value).push_back(bp::extract<int>(tmplist[j]));	    
	    (*frompy)[bp::extract<int>(keys[i])] = value;
	  }
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "map_vector_float"){
	  bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
	  bp::list keys = bp::extract<bp::list>(map.attr("keys")());
	  RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
	  for(int i=0; i<bp::len(keys); i++){
	    RCPtr<Vector<float> > value = 
	      RCPtr<Vector<float> >(Vector<float>::alloc(bp::len(keys)));
	    bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
	    for(int j=0; j<bp::len(tmplist); j++)
	      (*value).push_back(bp::extract<float>(tmplist[j]));
	    (*frompy)[bp::extract<int>(keys[i])] = value;
	  }
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "map_vector_complex"){
	  bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
	  bp::list keys = bp::extract<bp::list>(map.attr("keys")());
	  RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
	  for(int i=0; i<bp::len(keys); i++){
	    RCPtr<Vector<complex<float> > > value = RCPtr<Vector<complex<float> > >(new Vector<complex<float> >);
	    bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
	    for(int j=0; j<bp::len(tmplist); j++)
	      (*value).push_back(bp::extract<complex<float> >(tmplist[j]));
	    (*frompy)[bp::extract<int>(keys[i])] = value;
	  }
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(*outputsTypeit == "map_prim_float"){
	  bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
	  bp::list keys = bp::extract<bp::list>(map.attr("keys")());
	  RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
	  for(int i=0; i<bp::len(keys); i++){
	    RCPtr<Float> value = RCPtr<Float>(new Float);
	    (*frompy)[bp::extract<int>(keys[i])] = value;
	  }
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
	}
	else if(outputsTypeit->find("map_string_map_string:", 0) != string::npos){
          unsigned long long int index, offset;
          map<unsigned long long int, string> typesMap;
          index = outputsTypeit->find("map_string_map_string:", 0) + 22;
          offset = index;
          while(offset != string::npos){
            offset = outputsTypeit->find("prim_int", offset);
            if(offset != string::npos){
              typesMap[offset]="prim_int";
              offset+=8;
            }
            //cerr << "prim_int: count" << count << " offset" << offset << " npos" << string::npos << endl;
          }
          offset = index;
          while(offset != string::npos){
            offset = outputsTypeit->find("prim_float", offset);
            if(offset != string::npos){
              typesMap[offset]="prim_float";
              offset+=10;
            }
            //cerr << "prim_float: count" << count << " offset" << offset << " npos" << string::npos << endl;
          }
cerr << "AAAAAAAAAAA" << endl;
          map<unsigned long long int, string>::iterator it = typesMap.begin();
          //do{
          //  unsigned int mapend = outputsTypeit->find(";", index);
          //  index+=22;
          //  if(mapend == string::npos){}
          //  else{}
          //} while((index = outputsTypeit->find("map_string_map_string:", index)) != string::npos);
cerr << "AAAAAAAAAAA0" << endl;
	  bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
	  bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          keys.attr("sort")();
cerr << "AAAAAAAAAAA1" << endl;
	  RCPtr<Map<string, Map<string, ObjectRef> > > frompy = RCPtr<Map<string, Map<string, ObjectRef> > >(new Map<string, Map<string, ObjectRef> >);
	  for(int i=0; i<bp::len(keys); i++){
	    RCPtr<Map<string, ObjectRef> > mapinmap = RCPtr<Map<string, ObjectRef> >(new Map<string, ObjectRef>);
	    bp::dict mapmap = bp::extract<bp::dict>(map.get(keys[i]));
	    bp::list mapkeys = bp::extract<bp::list>(mapmap.attr("keys")());
            mapkeys.attr("sort")();
cerr << "AAAAAAAAAAA2" << endl;
	    for(int j=0; j<bp::len(mapkeys); j++){
              if(it->second == "prim_int"){
cerr << "AAAAAAAAAAA3" << endl;
                RCPtr<Int> value = RCPtr<Int>(new Int);
                (*mapinmap)[bp::extract<string>(mapkeys[j])] = value;
              }
              else if(it->second == "prim_float"){
cerr << "AAAAAAAAAAA4" << endl;
                RCPtr<Float> value = RCPtr<Float>(new Float);
                (*mapinmap)[bp::extract<string>(mapkeys[j])] = value;
              }
cerr << "AAAAAAAAAAA5" << endl;
              it++;
            }
cerr << "AAAAAAAAAAA6" << endl;
	    (*frompy)[bp::extract<string>(keys[i])] = *mapinmap;
	  }
cerr << "AAAAAAAAAAA7" << endl;
	  (*(outputs[*outputsIDit].buffer))[count] = frompy;
cerr << "AAAAAAAAAAA8" << endl;
	}
	else{
	  stringstream ss;
	  ss << "Unsupported type for output:" << *outputsTypeit;
	  throw new NodeException (NULL, ss.str(), __FILE__, __LINE__);
	}
      }
    } catch(bp::error_already_set){
      PyErr_Print(); 
      stringstream ss;
      ss << "Error during setting outputs of type (" << *outputsTypeit << ") "
	 << "and name (" << *outputsNameit << ")" << endl;
      throw (new NodeException(NULL, ss.str(), __FILE__, __LINE__));      
    }

    isCalculated = true;
  }
};
